/*
Connect4 game implementation (platform agnostic).
*/

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h> // for rand()
#include <string.h> // for memcpy()
#include "game.h"
#include "io.h"

/// <summary>
/// The game board.
/// </summary>
char board[ROWS][COLS];
/// <summary>
/// Whether theplayer is making the 1st move.
/// </summary>
bool firstMove;
/// <summary>
/// Player types (human, computer)
/// </summary>
uint8_t player_types[2] = { PLAYER_TYPE_COMPUTER, PLAYER_TYPE_COMPUTER };
/// <summary>
/// Draw games counter (for statistics).
/// </summary>
uint16_t drawGames = 0;
/// <summary>
/// Wins counters by player (for statistics).
/// </summary>
uint16_t wins[2] = { 0, 0 };
/// <summary>
/// Total moves counter (for statistics).
/// </summary>
uint32_t totalMoves = 0;

// Human play function
uint8_t humanPlay(uint8_t player);
// Computer play fuction
uint8_t computerPlay(uint8_t player);
// Pointers to the human/computer play functions. The proper function will be called by indexing this array with the player type.
uint8_t(*playFunction[2]) (uint8_t player) = { humanPlay, computerPlay };

/// <summary>
/// Resets the game board, by setting each cell value to the empty character.
/// </summary>
void resetBoard(void)
{
	for (uint8_t r = 0; r < ROWS; r++)
	{
		for (uint8_t c = 0; c < COLS; c++)
		{
			board[r][c] = EMPTY;
		}
	}
}

/// <summary>
/// Drops a disc into the board for the provided player at the specified column number.
/// </summary>
/// <param name="board">The board.</param>
/// <param name="player">Th eindex (0 or 1) of the player making the move.</param>
/// <param name="column">The selected column number.</param>
/// <returns>The index of the row in which the disc stops.</returns>
uint8_t dropDisc(char board[ROWS][COLS], int player, int column)
{
	for (int8_t r = ROWS - 1; r >= 0; r--)
	{
		if (board[r][column] == EMPTY)
		{
			board[r][column] = '0' + player;
			return r;
		}
	}

	// Should never go here
	return 255;
}

/// <summary>
/// Scans the board by checking for horizontal, vertical or diagonal lines lineLen long.
/// The function immediately returns when a line is found, so must be called after every move.
/// </summary>
/// <param name="board">The game board.</param>
/// <param name="lineLen">The length of the lines to check.</param>
/// <returns>Returns the game status:
/// 0: player 0 wins (the first line found is made up of '0's)
/// 1: player 1 wins (the first line found is made up of '1's)
/// 2: draw game (no line is found and the board is full)
/// 3: game not over (no line is found and the board is not full)
/// </returns>
uint8_t checkGame(char board[ROWS][COLS], uint8_t lineLen)
{
	uint8_t i;
	uint8_t result = GAME_STATUS_DRAW;

	for (uint8_t r = 0; r < ROWS; r++)
	{
		for (uint8_t c = 0; c < COLS; c++)
		{
			char ch = board[r][c];
			if (ch == EMPTY)
			{
				result = GAME_STATUS_NOT_OVER;
				continue;
			}

			// Horizontal line check
			if (c <= COLS - lineLen)
			{
				for (i = c + 1; i < c + lineLen; i++)
				{
					if (board[r][i] != ch)
						break;
				}
				if (i == c + lineLen)
				{
					// Horizontal line
					return ch - '0';
				}
			}

			// Vertical line check
			if (r <= ROWS - lineLen)
			{
				for (i = r + 1; i < r + lineLen; i++)
				{
					if (board[i][c] != ch)
						break;
				}
				if (i == r + lineLen)
				{
					// Vertical line
					return ch - '0';
				}
			}

			// Diagonal \ line check
			if (c <= COLS - lineLen && r <= ROWS - lineLen)
			{
				for (i = 1; i < lineLen; i++)
				{
					if (board[r + i][c + i] != ch)
						break;
				}
				if (i == lineLen)
				{
					return ch - '0';
				}
			}

			// Diagonal / line check
			if (c <= COLS - lineLen && r >= lineLen - 1)
			{
				for (i = 1; i < lineLen; i++)
				{
					if (board[r - i][c + i] != ch)
						break;
				}
				if (i == lineLen)
				{
					return ch - '0';
				}
			}

		}
	}
	return result;
}

/// <summary>
/// Executes a move.
/// </summary>
/// <param name="player">The index of the computer player (0 or 1).</param>
/// <returns>The column selected by the computer player.</returns>
uint8_t computerPlay(uint8_t player)
{
	int8_t c, r;
	uint8_t status;
	static char tempBoard[ROWS][COLS];
	static char tempBoard2[ROWS][COLS];
	uint8_t availableCols[COLS];
	uint8_t availableColsCount;
	uint8_t opponent = 1 - player;

	printThinking(player);

	if (firstMove)
	{
		// Drop disc in the middle column.
		return COLS >> 1;
	}

	// Check if can win
	for (c = 0; c < COLS; c++)
	{
		memcpy(tempBoard, board, ROWS * COLS * sizeof(char));

		if (tempBoard[0][c] == EMPTY)
		{
			dropDisc(tempBoard, player, c);
			status = checkGame(tempBoard, LINELEN);
			if (status == player)
				return c;
		}
	}

	// Check if can stop the opponent player from winning
	for (c = 0; c < COLS; c++)
	{
		memcpy(tempBoard, board, ROWS * COLS * sizeof(char));

		if (tempBoard[0][c] == EMPTY)
		{
			dropDisc(tempBoard, opponent, c);
			status = checkGame(tempBoard, LINELEN);
			if (status == opponent)
				return c;
		}
	}

	// Avoid moves that could lead the opponent to winning the game
	availableColsCount = 0;
	for (c = 0; c < COLS; c++)
	{
		memcpy(tempBoard, board, ROWS * COLS * sizeof(char));

		if (tempBoard[0][c] == EMPTY)
		{
			dropDisc(tempBoard, player, c);
			
			bool opponentCanWin = false;

			for (uint8_t oc = 0; oc < COLS; oc++)
			{
				memcpy(tempBoard2, tempBoard, ROWS * COLS * sizeof(char));

				if (tempBoard2[0][oc] == EMPTY)
				{
					dropDisc(tempBoard2, opponent, oc);

					status = checkGame(tempBoard2, LINELEN);
					if (status == opponent)
					{
						opponentCanWin = true;
						break;
					}
				}
			}

			if (!opponentCanWin)
			{
				availableCols[availableColsCount++] = c;
			}
		}
	}
	
	if (availableColsCount == 0)
	{
		for (c = 0; c < COLS; c++)
		{
			if (board[0][c] == EMPTY)
				availableCols[availableColsCount++] = c;
		}
	}
	

	// Prefer center columns
	if (rand() % 256 > 127)
	{
		for (c = (availableColsCount >> 1); c < availableColsCount; c++)
		{
			if (rand() % 256 > 95)
			{
				return availableCols[c];
			}
		}
	}
	else
	{
		for (c = (availableColsCount >> 1); c >= 0; c--)
		{
			if (rand() % 256 > 95)
			{
				return availableCols[c];
			}
		}
	}

	c = rand() % availableColsCount;
	return availableCols[c];
}

/// <summary>
/// Prompts the human player for a valid column number.
/// </summary>
/// <param name="player">The index of the human player (0 or 1).</param>
/// <returns>The column selected by the human player.</returns>
uint8_t humanPlay(uint8_t player)
{
	uint8_t selectedCol;
	do
	{
		selectedCol = promptColumn(player);
	} 			
	while (board[0][selectedCol] != EMPTY);
	return selectedCol;
}

/// <summary>
/// The main game loop:
/// * Check board status
///		* If game is over (either when game ends in a draw or one of the two players win)
///			* Return
///		* If game is not over
///			* Wait for current player's move
///			* Execute player move by dropping its disc in the selected column
///			* Change players turn
/// </summary>
/// <param name="firstPlayer">The index of the player starting the game (0 or 1).</param>
void game(uint8_t firstPlayer)
{
	uint8_t player, status;
	uint8_t moves = 0;

	resetBoard();
	firstMove = true;
	player = firstPlayer;

	printBoard();

	// Loop
	while (true)
	{
		status = checkGame(board, LINELEN);
		switch (status)
		{
		case GAME_STATUS_DRAW:
			printDraw();
			++drawGames;
			return;
		case GAME_STATUS_NOT_OVER:
			break;
		default:
			printWinner(status, moves);
			wins[status]++;
			return;
		}

		uint8_t selectedCol = playFunction[player_types[player]](player);
		printMove(player, selectedCol);

		uint8_t stopRow = dropDisc(board, player, selectedCol);
		updateBoard(player, selectedCol, stopRow);
		if (player == firstPlayer)
		{
			++moves;
			++totalMoves;
		}
		firstMove = false;

		player = 1 - player;
	}
}
