/*
Connect4 I/O and UI functions declarations, that must be implemented for each platform.

Copyright (c) 2021 Marco's Retrobits

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#pragma once

#include <stdint.h>

/// <summary>
/// Platform, UI and I/O initialization.
/// </summary>
void init(void);

/// <summary>
/// Displays the game statistics and the main menu and waits for the user's choice.
/// The available choices are:
/// 1: start a new match
/// 2: change player 0 type (human or computer)
/// 3: change player 1 type (human or computer)
/// 4: change the player making the first move in the next match
/// 5: resets game statistics
/// </summary>
/// <param name="firstPlayer">The index (0 or 1) of the player that will make the first move in the next match.</param>
uint8_t doMenu(uint8_t firstPlayer);

/// <summary>
/// Displays the game board.
/// </summary>
void printBoard(void);

/// <summary>
/// Updates the game board with the last move.
/// </summary>
/// <param name="player">The index of the player who made the last move.</param>
/// <param name="col">The last move (selected column).</param>
/// <param name="row">The row at which the disc stops.</param>
void updateBoard(uint8_t player, uint8_t col, uint8_t row);

/// <summary>
/// Prompts the player for his move (select column in which he will drop the disc).
/// </summary>
/// <param name="firstPlayer">The index (0 or 1) of the player.</param>
uint8_t promptColumn(uint8_t player);

/// <summary>
/// Displays the player winning the match.
/// </summary>
/// <param name="winner">The index (0 or 1) of the player winning the match.</param>
/// <param name="moves">How many moves the player made to win the match.</param>
void printWinner(uint8_t winner, uint8_t moves);

/// <summary>
/// Displays the draw game message. 
/// </summary>
void printDraw(void);

/// <summary>
/// Displays a message informing about the move chosen by the player.
/// </summary>
/// <param name="player">The index of the player who made the last move.</param>
/// <param name="col">The last move (selected column).</param>
void printMove(uint8_t player, uint8_t col);

/// <summary>
/// Displays a message informing that the player controlled by the computer is deciding its next move.
/// </summary>
/// <param name="player">The index of the player who is currently thinking.</param>
void printThinking(uint8_t player);
